import _ from "lodash";

import { Chess } from "../chessops/src/chess";
import { Move } from "../chessops/src/types";

import {
  Atomic,
  Antichess,
  Crazyhouse,
  Horde,
  KingOfTheHill,
  RacingKings,
  ThreeCheck,
} from "../chessops/src/variant";

import { parseSan, makeSan } from "../chessops/src/san";

import { makeFen, parseFen } from "../chessops/src/fen";

import { parseUci, makeUci } from "../chessops/src/util";

import { chessgroundDests } from "../chessops/src/compat";

export { chessgroundDests };

export type Uci = string;
export type San = string;
export type Variant = string;
export type Fen = string;

const PROMOTION_PIECES = ["q", "r", "b", "n", "k"];
const PROMOTION_PIECES_EXT = [""].concat(PROMOTION_PIECES);

// Pos_ is an abstraction of a chess position
export class Pos_ {
  pos: Chess;

  constructor() {
    // initialize to standard chess starting position
    this.pos = Chess.default();
  }

  get rules(): Variant {
    return this.pos.rules;
  }

  get variant(): Variant {
    return this.rules;
  }

  rawLegalUcis() {
    return Array.from(chessgroundDests(this.pos).entries())
      .map((entry) => entry[1].map((dest) => `${entry[0]}${dest}`))
      .flat();
  }

  setVariant(variant: Variant) {
    switch (variant) {
      case "atomic":
        this.pos = Atomic.default();
        break;
      case "antichess":
        this.pos = Antichess.default();
        break;
      case "crazyhouse":
        this.pos = Crazyhouse.default();
        break;
      case "horde":
        this.pos = Horde.default();
        break;
      case "kingofthehill":
        this.pos = KingOfTheHill.default();
        break;
      case "racingkings":
        this.pos = RacingKings.default();
        break;
      case "3check":
      case "threecheck":
        this.pos = ThreeCheck.default();
        break;
      default:
        this.pos = Chess.default();
    }
    return this;
  }

  setFen(fen: Fen) {
    const setupResult = parseFen(fen);

    if (setupResult.isErr) {
      return this;
    }

    const setup = setupResult.value;

    let posResult = undefined;

    const variant = this.pos.rules;

    switch (variant) {
      case "atomic":
        posResult = Atomic.fromSetup(setup);
        break;
      case "antichess":
        posResult = Antichess.fromSetup(setup);
        break;
      case "crazyhouse":
        posResult = Crazyhouse.fromSetup(setup);
        break;
      case "horde":
        posResult = Horde.fromSetup(setup);
        break;
      case "kingofthehill":
        posResult = KingOfTheHill.fromSetup(setup);
        break;
      case "racingkings":
        posResult = RacingKings.fromSetup(setup);
        break;
      case "3check":
        posResult = ThreeCheck.fromSetup(setup);
        break;
      default:
        posResult = Chess.fromSetup(setup);
    }

    if (posResult === undefined) return this;

    if (posResult.isErr) return this;

    this.pos = posResult.value;

    return this;
  }

  reportFen() {
    return makeFen(this.pos.toSetup());
  }

  sanToMove(san: San): Move | undefined {
    return parseSan(this.pos, san);
  }

  moveToSan(move: Move) {
    return makeSan(this.pos, move);
  }

  uciToMove(uci: Uci): Move | undefined {
    const move = parseUci(uci);
    return move;
  }

  moveToUci(move: Move) {
    return makeUci(move);
  }

  play(move: Move) {
    this.pos.play(move);
    return this;
  }

  playSan(san: San) {
    const move = this.sanToMove(san);
    if (move === undefined) return this;
    return this.play(move);
  }

  playUci(uci: Uci) {
    const move = this.uciToMove(uci);
    if (move === undefined) return this;
    return this.play(move);
  }

  sanToUci(san: San): Uci | undefined {
    const move = this.sanToMove(san);
    if (move === undefined) return undefined;
    return this.moveToUci(move);
  }

  uciToSan(uci: Uci): San | undefined {
    const move = this.uciToMove(uci);
    if (move === undefined) return undefined;
    return this.moveToSan(move);
  }

  toString() {
    return `[Pos ${this.pos.rules} ${this.reportFen()}]`;
  }

  legalsForUci(uci: Uci) {
    return PROMOTION_PIECES_EXT.filter((pp) => {
      const move = this.uciToMove(`${uci}${pp}`);
      if (move === undefined) return false;
      const legal = this.pos.isLegal(move, this.pos.ctx());
      return legal;
    }).map((pp) => `${uci}${pp}`);
  }

  allLegalUcis() {
    return this.rawLegalUcis()
      .map((uci) => this.legalsForUci(uci))
      .flat();
  }

  allLegalSans() {
    const rawLegalSans = this.allLegalUcis().map((uci) => this.uciToSan(uci));
    const legalSans = _.uniq(rawLegalSans);
    return legalSans;
  }
}

export function Pos() {
  return new Pos_();
}
